// dart_mssql.cpp : Defines the exported functions for the DLL application.
//
#include <comdef.h>     
#include <comutil.h>
#include <map>
#include <list>
#include <exception>
#include "msoledbsql.h"
#include "json.hpp"
#include "dart_api.h"
#include "dart_native_api.h"
#include "dart_mssql.h"

Dart_Handle HandleError(Dart_Handle handle) {
	if (Dart_IsError(handle)) Dart_PropagateError(handle);
	return handle;
}

class classExecute {
private:
	int	 errorCount = 0;
	std::string errorMessages;
public:
	classExecute(Dart_Port reply_port_id, char* serverName, char* dbName, char* userName, char* password, int authType, char* sqlCommand, Dart_CObject* result) {
		CoInitializeEx(NULL, COINIT_MULTITHREADED);
		IDBInitialize* pSession = NULL;
		json jsonResult;
		result->type = Dart_CObject_kString;
		sqlConnect(serverName, dbName, userName, password, authType, &pSession, &errorCount, &errorMessages);
		if (errorCount == 0) {
			BSTR sqlCmd = UTF8ToBSTR(sqlCommand);
			jsonResult = sqlExecute(pSession, sqlCmd, &errorCount, &errorMessages);
			SysFreeString(sqlCmd);
			if (jsonResult["rowsAffected"] != -1) {
				char s1[20] = "select @@identity";
				BSTR s2 = UTF8ToBSTR(s1);
				json lastIdentity = sqlExecute(pSession, s2, &errorCount, &errorMessages);
				jsonResult["lastIdentity"] = lastIdentity["rows"][0][0];
				SysFreeString(s2);
			}
		}
		if (errorCount != 0) {
			jsonResult["error"] = errorMessages;
		}
		std::string dumped = jsonResult.dump();
		result->value.as_string = new char[dumped.length()+1];
		strcpy_s(result->value.as_string, dumped.length()+1, dumped.c_str());
		Dart_PostCObject(reply_port_id, result);
		delete[] result->value.as_string;
		if (pSession != NULL) {
			pSession->Uninitialize();
			pSession->Release();
		}
	}
	~classExecute() {
		CoUninitialize();
	}
};

void wrappedExecute(Dart_Port dest_port_id, Dart_CObject* message) {
	Dart_Port reply_port_id = ILLEGAL_PORT;
	if (message->type == Dart_CObject_kArray && message->value.as_array.length == 7) {
		Dart_CObject* param0 = message->value.as_array.values[0];
		Dart_CObject* param1 = message->value.as_array.values[1];
		Dart_CObject* param2 = message->value.as_array.values[2];
		Dart_CObject* param3 = message->value.as_array.values[3];
		Dart_CObject* param4 = message->value.as_array.values[4];
		Dart_CObject* param5 = message->value.as_array.values[5];
		Dart_CObject* param6 = message->value.as_array.values[6];
		if (param0->type == Dart_CObject_kSendPort) {
			Dart_CObject result;
			reply_port_id = param0->value.as_send_port.id;
			if (param1->type == Dart_CObject_kString &&
				param2->type == Dart_CObject_kString &&
				param3->type == Dart_CObject_kString &&
				param4->type == Dart_CObject_kString &&
				param6->type == Dart_CObject_kString) {
				int authType = 0; // user + senha
				if (param5->type == Dart_CObject_kInt32) {
					authType = (int)param5->value.as_int32;
				}
				classExecute execute(reply_port_id, param1->value.as_string, param2->value.as_string, param3->value.as_string, param4->value.as_string, authType, param6->value.as_string, &result);
			}
			else {
				result.type = Dart_CObject_kNull;
				Dart_PostCObject(reply_port_id, &result);
			}
		}
	}
}

void executeServicePort(Dart_NativeArguments arguments) {
	Dart_EnterScope();
	Dart_SetReturnValue(arguments, Dart_Null());
	Dart_Port service_port =
		Dart_NewNativePort("executeService", wrappedExecute, true);
	if (service_port != ILLEGAL_PORT) {
		Dart_Handle send_port = HandleError(Dart_NewSendPort(service_port));
		Dart_SetReturnValue(arguments, send_port);
	}
	Dart_ExitScope();
}

struct FunctionLookup {
	const char* name;
	Dart_NativeFunction function;
};

FunctionLookup function_list[] = {
	{"execute_ServicePort", executeServicePort},
	{NULL, NULL}
};

Dart_NativeFunction nativeResolver(Dart_Handle name, int argc, bool * autoSetupScope) {
	if (!Dart_IsString(name)) {
		return NULL;
	}
	Dart_NativeFunction result = NULL;
	if (autoSetupScope == NULL) {
		return NULL;
	}

	Dart_EnterScope();
	const char* cname;
	HandleError(Dart_StringToCString(name, &cname));

	for (int i = 0; function_list[i].name != NULL; ++i) {
		if (strcmp(function_list[i].name, cname) == 0) {
			*autoSetupScope = true;
			result = function_list[i].function;
			break;
		}
	}

	if (result != NULL) {
		Dart_ExitScope();
		return result;
	}

	Dart_ExitScope();
	return result;
}

DART_EXPORT Dart_Handle dart_mssql_Init(Dart_Handle library) {
	if (Dart_IsError(library)) {
		return library;
	}
	Dart_Handle result = Dart_SetNativeResolver(library, nativeResolver, NULL);
	if (Dart_IsError(result)) {
		return result;
	}
	return Dart_Null();
}

