#include <stdio.h>
#include <comutil.h>
#include "stdafx.h"
#include "msoledbsql.h"
#include "json.hpp"

//ROUNDUP on all platforms pointers must be aligned properly
#define ROUNDUP_AMOUNT	8
#define ROUNDUP_(size,amount)	(((DBBYTEOFFSET)(size)+((amount)-1))&~((DBBYTEOFFSET)(amount)-1))
#define ROUNDUP(size)			ROUNDUP_(size, ROUNDUP_AMOUNT)

BSTR intToBSTR(int number) {
	wchar_t buf[32];
	int len = swprintf(buf, 32, L"%d", number);
	BSTR result = SysAllocStringLen(buf, len);
	return result;
}

void addError(int* errorCount, std::string* errorMessages, std::string errorMessage) {
	(*errorMessages).append(errorMessage);
	(*errorCount)++;
}

BSTR UTF8ToBSTR(char* utf8Value) {
	int wslen = MultiByteToWideChar(CP_UTF8, 0, utf8Value, (int)strlen(utf8Value), 0, 0);
	BSTR result = SysAllocStringLen(0, wslen);
	MultiByteToWideChar(CP_UTF8, 0, utf8Value, (int)strlen(utf8Value), result, wslen);
	return result;
}

std::string WideCharToUTF8(const wchar_t* value) {
	if (value == NULL) {
		return NULL;
	}
	int n = WideCharToMultiByte(CP_UTF8, 0, value, -1, NULL, 0, NULL, NULL);
	if (n <= 0) {
		return NULL;
	}
	char* buffer = new char[n];
	WideCharToMultiByte(CP_UTF8, 0, value, -1, buffer, n, NULL, NULL);
	std::string str;
	str.assign(buffer);
	delete[] buffer;
	return str;
}

HRESULT oleCheck(HRESULT hr, int* errorCount, std::string* errorMessages) {
	// fill errorMessages if last operation had a error 
	if (FAILED(hr) && errorCount != NULL) {
		IErrorInfo		*pIErrorInfo = NULL;
		IErrorRecords	*pIErrorRecords;

		(*errorMessages).assign("");
		GetErrorInfo(0, &pIErrorInfo);
		if (pIErrorInfo != NULL && *errorCount == 0) {
			HRESULT hr1 = pIErrorInfo->QueryInterface(IID_IErrorRecords, (void **)&pIErrorRecords);
			if (SUCCEEDED(hr1)) {
				ULONG pcRecords;
				pIErrorRecords->GetRecordCount(&pcRecords);
				*errorCount = pcRecords;
				boolean ok = FALSE;
				for (ULONG i = 0; i < pcRecords; i++) {
					ISQLServerErrorInfo *pISQLServerErrorInfo;
					pIErrorRecords->GetCustomErrorObject(
						i, IID_ISQLServerErrorInfo,
						(IUnknown**)&pISQLServerErrorInfo);
					ok = (pISQLServerErrorInfo > 0);
					if (ok) {
						SSERRORINFO *errorInfo;
						OLECHAR *errorMsgs;
						pISQLServerErrorInfo->GetErrorInfo(&errorInfo, &errorMsgs);
						ok = (errorInfo != NULL);
						if (ok) {
							(*errorMessages).append(WideCharToUTF8(errorMsgs));
							CoTaskMemFree(errorInfo);
							CoTaskMemFree(errorMsgs);
						}
						pISQLServerErrorInfo->Release();
					}
					if (!ok) {
						OLECHAR *errorMsgs;
						pIErrorRecords->GetErrorInfo(i, 0, &pIErrorInfo);
						pIErrorInfo->GetDescription(&errorMsgs);
						ok = (errorMsgs != NULL);
						if (ok) {
							(*errorMessages).append(WideCharToUTF8(errorMsgs));
							pIErrorInfo->Release();
							SysFreeString(errorMsgs);
						}
					}
				}
				pIErrorRecords->Release();
			}
		}
		if ((*errorMessages).length() == 0) {
			char errorBuf[32];
			int len = snprintf(errorBuf, 32, "error code: 0x%X\0", hr);
			addError(errorCount, errorMessages, errorBuf);
		}
		if (pIErrorInfo != NULL) {
			pIErrorInfo->Release();
		}
	}
	return hr;
}

HRESULT memoryCheck(HRESULT hr, void *pv, int* errorCount, std::string* errorMessages) {
	if (!pv) {
		hr = E_OUTOFMEMORY;
		return oleCheck(hr, errorCount, errorMessages);
	}
	else {
		return hr;
	}
}

/////////////////////////////////////////////////////////////////
// freeBindings
//
//	This function frees a bindings array and any allocated
//	structures contained in that array.
//
/////////////////////////////////////////////////////////////////
void freeBindings(DBORDINAL	cBindings, DBBINDING* rgBindings) {
	ULONG iBind;

	// Free any memory used by DBOBJECT structures in the array
	for (iBind = 0; iBind < cBindings; iBind++)
		CoTaskMemFree(rgBindings[iBind].pObject);

	// Now free the bindings array itself
	CoTaskMemFree(rgBindings);
}

/////////////////////////////////////////////////////////////////
// getProperty
//
//	This function gets the BOOL value for the specified property
//	and returns the result in *pbValue.
//
//  here we do not pass errorCount and errorMessages since errors are expected for non-supported properties
/////////////////////////////////////////////////////////////////
HRESULT getProperty(IUnknown* pIUnknown, REFIID	riid, DBPROPID dwPropertyID, REFGUID guidPropertySet, BOOL* pbValue) {
	HRESULT					hr;
	DBPROPID				rgPropertyIDs[1];
	DBPROPIDSET				rgPropertyIDSets[1];

	ULONG					cPropSets = 0;
	DBPROPSET*				rgPropSets = NULL;

	IDBProperties*			pIDBProperties = NULL;
	ISessionProperties*		pISesProps = NULL;
	ICommandProperties*		pICmdProps = NULL;
	IRowsetInfo*			pIRowsetInfo = NULL;

	// Initialize the output value
	*pbValue = FALSE;

	// Set up the property ID array
	rgPropertyIDs[0] = dwPropertyID;

	// Set up the Property ID Set
	rgPropertyIDSets[0].rgPropertyIDs = rgPropertyIDs;
	rgPropertyIDSets[0].cPropertyIDs = 1;
	rgPropertyIDSets[0].guidPropertySet = guidPropertySet;

	// Get the property value for this property from the provider, but
	// don't try to display extended error information, since this may
	// not be a supported property: a failure is, in fact, expected if
	// the property is not supported
	if (riid == IID_IDBProperties) {
		hr = pIUnknown->QueryInterface(IID_IDBProperties,
			(void**)&pIDBProperties);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
		hr = pIDBProperties->GetProperties(
			1,					//cPropertyIDSets
			rgPropertyIDSets,	//rgPropertyIDSets
			&cPropSets,			//pcPropSets
			&rgPropSets			//prgPropSets);
		);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
	}
	else if (riid == IID_ISessionProperties) {
		hr = pIUnknown->QueryInterface(IID_ISessionProperties,
			(void**)&pISesProps);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
		hr = pISesProps->GetProperties(
			1,					//cPropertyIDSets
			rgPropertyIDSets,	//rgPropertyIDSets
			&cPropSets,			//pcPropSets
			&rgPropSets			//prgPropSets
		);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
	}
	else if (riid == IID_ICommandProperties) {
		hr = pIUnknown->QueryInterface(IID_ICommandProperties,
			(void**)&pICmdProps);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
		hr = pICmdProps->GetProperties(
			1,					//cPropertyIDSets
			rgPropertyIDSets,	//rgPropertyIDSets
			&cPropSets,			//pcPropSets
			&rgPropSets			//prgPropSets
		);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
	}
	else {
		hr = pIUnknown->QueryInterface(IID_IRowsetInfo,
			(void**)&pIRowsetInfo);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
		hr = pIRowsetInfo->GetProperties(
			1,					//cPropertyIDSets
			rgPropertyIDSets,	//rgPropertyIDSets
			&cPropSets,			//pcPropSets
			&rgPropSets			//prgPropSets
		);
		if (oleCheck(hr, NULL, NULL) < 0) goto CLEANUP;
	}

	// Return the value for this property to the caller if
	// it's a VT_BOOL type value, as expected
	if (V_VT(&rgPropSets[0].rgProperties[0].vValue) == VT_BOOL)
		*pbValue = V_BOOL(&rgPropSets[0].rgProperties[0].vValue);

CLEANUP:
	if (rgPropSets) {
		CoTaskMemFree(rgPropSets[0].rgProperties);
		CoTaskMemFree(rgPropSets);
	}
	if (pIDBProperties)
		pIDBProperties->Release();
	if (pISesProps)
		pISesProps->Release();
	if (pICmdProps)
		pICmdProps->Release();
	if (pIRowsetInfo)
		pIRowsetInfo->Release();
	return hr;
}

/////////////////////////////////////////////////////////////////
// setupBindings
//
//	This function takes an IUnknown pointer from a Rowset object
//	and creates a bindings array that describes how we want the
//	data we fetch from the Rowset to be laid out in memory. It
//	also calculates the total size of a row so that we can use
//	this to allocate memory for the rows that we will fetch
//	later.
//
//	For each column in the Rowset, there will be a corresponding
//	element in the bindings array that describes how the
//	provider should transfer the data, including length and
//	status, for that column. This element also specifies the data
//	type that the provider should return the column as. We will
//	bind all columns as DBTYPE_WSTR, with a few exceptions
//	detailed below, as providers are required to support the
//	conversion of their column data to this type in the vast
//	majority of cases. The exception to our binding as
//	DBTYPE_WSTR is if the native column data type is
//	DBTYPE_IUNKNOWN or if the user has requested that BLOB
//	columns be bound as ISequentialStream objects, in which case
//	we will bind those columns as ISequentialStream objects.
//
/////////////////////////////////////////////////////////////////
HRESULT setupBindings(IUnknown *pUnkRowset, DBORDINAL *pcBindings, DBBINDING **prgBindings, DBCOLUMNINFO **prgColumnInfo, LPWSTR* ppStringBuffer, DBORDINAL *pcbRowSize, int* errorCount, std::string* errorMessages) {
	HRESULT					hr;
	DBORDINAL				cColumns;	
	LPWSTR					pStringBuffer = NULL;
	IColumnsInfo*			pIColumnsInfo = NULL;

	DBLENGTH				iCol;
	DBORDINAL				dwOffset = 0;
	DBBINDING*				rgBindings = NULL;
	DBCOLUMNINFO*			rgColumnInfo = NULL;

	DBLENGTH				cStorageObjs = 0;
	BOOL					fMultipleObjs = FALSE;

	// Obtain the column information for the Rowset; from this, we can find
	// out the following information that we need to construct the bindings
	// array:
	//  - the number of columns
	//  - the ordinal of each column
	//	- the precision and scale of numeric columns
	//  - the OLE DB data type of the column
	hr = pUnkRowset->QueryInterface(
		IID_IColumnsInfo, (void**)&pIColumnsInfo);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;
	hr = pIColumnsInfo->GetColumnInfo(
		&cColumns,		//pcColumns
		&rgColumnInfo,	//prgColumnInfo
		&pStringBuffer	//ppStringBuffer
	);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	// Allocate memory for the bindings array; there is a one-to-one
	// mapping between the columns returned from GetColumnInfo and our
	// bindings
	rgBindings = (DBBINDING*)CoTaskMemAlloc(cColumns * sizeof(DBBINDING));
	if (memoryCheck(hr, rgBindings, errorCount, errorMessages) < 0) goto CLEANUP;
	memset(rgBindings, 0, cColumns * sizeof(DBBINDING));

	// Determine if the Rowset supports multiple storage object bindings;
	// if it does not, we will only bind the first BLOB column or IUnknown
	// column as an ISequentialStream object, and will bind the rest as
	// DBTYPE_WSTR
	getProperty(pUnkRowset, IID_IRowset, DBPROP_MULTIPLESTORAGEOBJECTS,
		DBPROPSET_ROWSET, &fMultipleObjs);

	// Construct the binding array element for each column
	for (iCol = 0; iCol < cColumns; iCol++) {
		// This binding applies to the ordinal of this column
		rgBindings[iCol].iOrdinal = rgColumnInfo[iCol].iOrdinal;

		// We are asking the provider to give us the data for this column
		// (DBPART_VALUE), the length of that data (DBPART_LENGTH), and
		// the status of the column (DBPART_STATUS)
		rgBindings[iCol].dwPart = DBPART_VALUE | DBPART_LENGTH | DBPART_STATUS;

		// The following values are the offsets to the status, length, and
		// data value that the provider will fill with the appropriate values
		// when we fetch data later. When we fetch data, we will pass a
		// pointer to a buffer that the provider will copy column data to,
		// in accordance with the binding we have provided for that column;
		// these are offsets into that future buffer
		rgBindings[iCol].obStatus = dwOffset;
		rgBindings[iCol].obLength = dwOffset + sizeof(DBSTATUS);
		rgBindings[iCol].obValue = dwOffset + sizeof(DBSTATUS) + sizeof(DBORDINAL);
		rgBindings[iCol].pTypeInfo = NULL;
		rgBindings[iCol].pBindExt = NULL;

		// Any memory allocated for the data value will be owned by us, the
		// client. Note that no data will be allocated in this case, as the
		// DBTYPE_WSTR bindings we are using will tell the provider to simply
		// copy data directly into our provided buffer
		rgBindings[iCol].dwMemOwner = DBMEMOWNER_CLIENTOWNED;

		// This is not a parameter binding
		rgBindings[iCol].eParamIO = DBPARAMIO_NOTPARAM;

		// We want to use the precision and scale of the column
		rgBindings[iCol].bPrecision = rgColumnInfo[iCol].bPrecision;
		rgBindings[iCol].bScale = rgColumnInfo[iCol].bScale;

		// Bind this column as DBTYPE_WSTR, which tells the provider to
		// copy a Unicode string representation of the data into our buffer,
		// converting from the native type if necessary
		rgBindings[iCol].wType = DBTYPE_WSTR;

		// Initially, we set the length for this data in our buffer to 0;
		// the correct value for this will be calculated directly below
		rgBindings[iCol].cbMaxLen = 0;

		// Determine the maximum number of bytes required in our buffer to
		// contain the Unicode string representation of the provider's native
		// data type, including room for the NULL-termination character
		switch (rgColumnInfo[iCol].wType)
		{
		case DBTYPE_NULL:
		case DBTYPE_EMPTY:
		case DBTYPE_I1:
		case DBTYPE_I2:
		case DBTYPE_I4:
		case DBTYPE_UI1:
		case DBTYPE_UI2:
		case DBTYPE_UI4:
		case DBTYPE_R4:
		case DBTYPE_BOOL:
		case DBTYPE_I8:
		case DBTYPE_UI8:
		case DBTYPE_R8:
		case DBTYPE_CY:
		case DBTYPE_ERROR:
			// When the above types are converted to a string, they
			// will all fit into 25 characters, so use that plus space
			// for the NULL-terminator
			rgBindings[iCol].cbMaxLen = (25 + 1) * sizeof(WCHAR);
			break;

		case DBTYPE_DECIMAL:
		case DBTYPE_NUMERIC:
		case DBTYPE_DATE:
		case DBTYPE_DBDATE:
		case DBTYPE_DBTIMESTAMP:
		case DBTYPE_GUID:
			// Converted to a string, the above types will all fit into
			// 50 characters, so use that plus space for the terminator
			rgBindings[iCol].cbMaxLen = (50 + 1) * sizeof(WCHAR);
			break;

		case DBTYPE_BYTES:
			// In converting DBTYPE_BYTES to a string, each byte
			// becomes two characters (e.g. 0xFF -> "FF"), so we
			// will use double the maximum size of the column plus
			// include space for the NULL-terminator
			rgBindings[iCol].cbMaxLen =
				(rgColumnInfo[iCol].ulColumnSize * 2 + 1) * sizeof(WCHAR);
			break;

		case DBTYPE_STR:
		case DBTYPE_WSTR:
		case DBTYPE_BSTR:
			// Going from a string to our string representation,
			// we can just take the maximum size of the column,
			// a count of characters, and include space for the
			// terminator, which is not included in the column size
			rgBindings[iCol].cbMaxLen =
				(rgColumnInfo[iCol].ulColumnSize + 1) * sizeof(WCHAR);
			break;

		default:
			// For any other type, we will simply use our maximum
			// column buffer size, since the display size of these
			// columns may be variable (e.g. DBTYPE_VARIANT) or
			// unknown (e.g. provider-specific types)
			if (*errorCount == 0) {
				addError(errorCount, errorMessages, "Unknown data type");
				hr = -1;
			}
			break;
		};

		// If the provider's native data type for this column is
		// DBTYPE_IUNKNOWN or this is a BLOB column and the user
		// has requested that we bind BLOB columns as ISequentialStream
		// objects, bind this column as an ISequentialStream object if
		// the provider supports our creating another ISequentialStream
		// binding
		if ((rgColumnInfo[iCol].wType == DBTYPE_IUNKNOWN ||
			((rgColumnInfo[iCol].dwFlags & DBCOLUMNFLAGS_ISLONG))) && (fMultipleObjs || !cStorageObjs)) {
			// To create an ISequentialStream object, we will
			// bind this column as DBTYPE_IUNKNOWN to indicate
			// that we are requesting this column as an object
			rgBindings[iCol].wType = DBTYPE_IUNKNOWN;

			// We want to allocate enough space in our buffer for
			// the ISequentialStream pointer we will obtain from
			// the provider
			rgBindings[iCol].cbMaxLen = sizeof(ISequentialStream *);

			// To specify the type of object that we want from the
			// provider, we need to create a DBOBJECT structure and
			// place it in our binding for this column
			rgBindings[iCol].pObject =
				(DBOBJECT *)CoTaskMemAlloc(sizeof(DBOBJECT));
			if (memoryCheck(hr, rgBindings[iCol].pObject, errorCount, errorMessages) < 0) goto CLEANUP;

			// Direct the provider to create an ISequentialStream
			// object over the data for this column
			rgBindings[iCol].pObject->iid = IID_ISequentialStream;

			// We want read access on the ISequentialStream
			// object that the provider will create for us
			rgBindings[iCol].pObject->dwFlags = STGM_READ;

			// Keep track of the number of storage objects (ISequentialStream
			// is a storage interface) that we have requested, so that we
			// can avoid requesting multiple storage objects from a provider
			// that only supports a single storage object in our bindings
			cStorageObjs++;
		}

		// Ensure that the bound maximum length is no more than the
		// maximum column size in bytes that we've defined
		// (comentado por moacir. Prefiro pegar tudo!) rgBindings[iCol].cbMaxLen = min(rgBindings[iCol].cbMaxLen, maxColSize);

		// Update the offset past the end of this column's data, so
		// that the next column will begin in the correct place in
		// the buffer
		dwOffset = rgBindings[iCol].cbMaxLen + rgBindings[iCol].obValue;

		// Ensure that the data for the next column will be correctly
		// aligned for all platforms, or, if we're done with columns,
		// that if we allocate space for multiple rows that the data
		// for every row is correctly aligned
		dwOffset = ROUNDUP(dwOffset);
	}

	// Return the row size (the current dwOffset is the size of the row),
	// the count of bindings, and the bindings array to the caller
	*pcbRowSize = dwOffset;
	*pcBindings = cColumns;
	*prgBindings = rgBindings;
	*prgColumnInfo = rgColumnInfo;
	*ppStringBuffer = pStringBuffer;

CLEANUP:
	if (pIColumnsInfo)
		pIColumnsInfo->Release();
	return hr;
}

/////////////////////////////////////////////////////////////////
// createAccessor
//
//	This function takes an IUnknown pointer for a Rowset object
//	and creates an Accessor that describes the layout of the
//	buffer we will use when we fetch data. The provider will fill
//	this buffer according to the description contained in the
//	Accessor that we will create here.
//
/////////////////////////////////////////////////////////////////
HRESULT createAccessor(IUnknown* pUnkRowset, HACCESSOR* phAccessor,	DBORDINAL* pcBindings, DBBINDING** prgBindings, DBCOLUMNINFO **prgColumnInfo, LPWSTR* ppStringBuffer, DBORDINAL* pcbRowSize, int* errorCount, std::string* errorMessages) {
	HRESULT					hr;
	IAccessor*				pIAccessor = NULL;

	// An Accessor is basically a handle to a collection of bindings.
	// To create the Accessor, we need to first create an array of
	// bindings for the columns in the Rowset
	hr = setupBindings(pUnkRowset, pcBindings, prgBindings, prgColumnInfo, ppStringBuffer, pcbRowSize, errorCount, errorMessages);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	// Now that we have an array of bindings, tell the provider to
	// create the Accessor for those bindings. We get back a handle
	// to this Accessor, which we will use when fetching data
	hr = pUnkRowset->QueryInterface(
		IID_IAccessor, (void**)&pIAccessor);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;
	hr = pIAccessor->CreateAccessor(
		DBACCESSOR_ROWDATA,	//dwAccessorFlags
		*pcBindings,		//cBindings
		*prgBindings,		//rgBindings
		0,					//cbRowSize
		phAccessor,			//phAccessor	
		NULL				//rgStatus
	);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;
CLEANUP:
	if (pIAccessor)
		pIAccessor->Release();
	return hr;
}

