#include <atlbase.h>
#include "msoledbsql.h"
#include "json.hpp"
#include "stdafx.h"
#include "oledberr.h"
#include "dart_api.h"
#include "dart_native_api.h"
#include "dart_mssql.h"

#define AUTH_TYPE_PASWORD		0
#define AUTH_TYPE_INTEGRATED	1

#define MAX_ROWS				1			// change from 10 to 1 since pIRowset->GetData does not work if pIRowset->GetNextRows returns more than one row
#define MAX_DISPLAY_SIZE		100

using namespace std;

HRESULT sqlConnect(char* serverName, char* dbName, char* userName, char* password, int authType, IDBInitialize** ppInitialize, int* errorCount, std::string* errorMessages) {
	IDBInitialize*		pIDBInitialize = nullptr;
	IDBProperties*		pIDBProperties = nullptr;
	DBPROP				InitProperties[4] = { 0 };
	DBPROPSET			rgInitPropSet[1] = { 0 };
	HRESULT				hr = S_OK;
	int					numProperties;

	*ppInitialize = NULL;
	
	// Obtain access to the OLE DB Driver for SQL Server.  
	hr = CoCreateInstance(CLSID_MSOLEDBSQL,
		NULL,
		CLSCTX_INPROC_SERVER,
		IID_IDBInitialize,
		(void **)&pIDBInitialize);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	// Initialize property values needed to establish connection.  
	for (int i = 0; i < 4; i++) {
		VariantInit(&InitProperties[i].vValue);
	}

	// Server name.  
	// See DBPROP structure for more information on InitProperties  
	InitProperties[0].dwPropertyID = DBPROP_INIT_DATASOURCE;
	InitProperties[0].vValue.vt = VT_BSTR;
	InitProperties[0].vValue.bstrVal = UTF8ToBSTR(serverName);
	InitProperties[0].dwOptions = DBPROPOPTIONS_REQUIRED;
	InitProperties[0].colid = DB_NULLID;

	// Database.  
	InitProperties[1].dwPropertyID = DBPROP_INIT_CATALOG;
	InitProperties[1].vValue.vt = VT_BSTR;
	InitProperties[1].vValue.bstrVal = UTF8ToBSTR(dbName);
	InitProperties[1].dwOptions = DBPROPOPTIONS_REQUIRED;
	InitProperties[1].colid = DB_NULLID;

	// Username (login).  
	if (authType == AUTH_TYPE_INTEGRATED) {
		InitProperties[2].dwPropertyID = DBPROP_AUTH_INTEGRATED;
		InitProperties[2].vValue.vt = VT_BSTR;
		InitProperties[2].vValue.bstrVal = SysAllocString(L"SSPI");
		InitProperties[2].dwOptions = DBPROPOPTIONS_REQUIRED;
		InitProperties[2].colid = DB_NULLID;
		numProperties = 3;
	}
	else {
		InitProperties[2].dwPropertyID = DBPROP_AUTH_USERID;
		InitProperties[2].vValue.vt = VT_BSTR;
		InitProperties[2].vValue.bstrVal = UTF8ToBSTR(userName);
		InitProperties[2].dwOptions = DBPROPOPTIONS_REQUIRED;
		InitProperties[2].colid = DB_NULLID;

		// password
		InitProperties[3].dwPropertyID = DBPROP_AUTH_PASSWORD;
		InitProperties[3].vValue.vt = VT_BSTR;
		InitProperties[3].vValue.bstrVal = UTF8ToBSTR(password);
		InitProperties[3].dwOptions = DBPROPOPTIONS_REQUIRED;
		InitProperties[3].colid = DB_NULLID;
		numProperties = 4;
	}

	// Construct the DBPROPSET structure(rgInitPropSet). The   
	// DBPROPSET structure is used to pass an array of DBPROP   
	// structures (InitProperties) to the SetProperties method.  
	rgInitPropSet[0].guidPropertySet = DBPROPSET_DBINIT;
	rgInitPropSet[0].cProperties = numProperties;
	rgInitPropSet[0].rgProperties = InitProperties;

	// Set initialization properties.  
	hr = pIDBInitialize->QueryInterface(IID_IDBProperties, (void **)&pIDBProperties);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	hr = pIDBProperties->SetProperties(1, rgInitPropSet);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	// Now establish the connection to the data source.  
	hr = pIDBInitialize->Initialize();
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	// I'm not sure if we need that
	for (int i = 0; i < 4; i++) {
		SysFreeString(InitProperties[i].vValue.bstrVal);
	}

	*ppInitialize = pIDBInitialize;

CLEANUP:
	if (pIDBProperties)	{
		pIDBProperties->Release();
		pIDBProperties = nullptr;
	}
	return hr;
}

json dbColumnValueToUF8(DBBINDING* rgBindings, DBCOLUMNINFO *rgColumnInfo, void* pData, ULONG iCol, int* errorCount, std::string* errorMessages) {
	DBSTATUS				dwStatus;
	ULONG					ulLength;
	void*					pvValue;
	ISequentialStream*		pISeqStream = NULL;
	ULONG					cbRead;
	char*					c = NULL;
	char*					longValueBuf = NULL;
	string					str;
	HRESULT					hr;
	BYTE					pbBuff[1000];
	ULONG					cbNeeded = sizeof(pbBuff) / sizeof(BYTE);
	ULONG					cbReadTotal = 0;
	IUnknown*				pIUnknown = NULL;

	// We have bound status, length, and the data value for all
	// columns, so we know that these can all be used
	dwStatus = *(DBSTATUS *)((BYTE *)pData + rgBindings[iCol].obStatus);
	ulLength = *(ULONG *)((BYTE *)pData + rgBindings[iCol].obLength);
	pvValue = (BYTE *)pData + rgBindings[iCol].obValue;
	json result;

	// Check the status of this column. This decides
	// exactly what will be displayed for the column
	switch (dwStatus) {
	// The data is NULL, so don't try to display it
	case DBSTATUS_S_ISNULL:
		result = nullptr;
		break;

	// The data was fetched, but may have been truncated.
	// Display string data for this column to the user
	case DBSTATUS_S_TRUNCATED:
	case DBSTATUS_S_OK:
	case DBSTATUS_S_DEFAULT:
		// We have either bound the column as a Unicode string
		// (DBTYPE_WSTR) or as an ISequentialStream object
		// (DBTYPE_IUNKNOWN), and have to do different processing
		// for each one of these possibilities
		switch (rgBindings[iCol].wType) {
		case DBTYPE_WSTR:
			switch (rgColumnInfo[iCol].wType) {
			case DBTYPE_NUMERIC:
			case DBTYPE_I4:
				result = wcstol((wchar_t *)pvValue, NULL, 10);
				break;
			default:
				// Copy the string data
				c = new char[ulLength * 2];
				WideCharToMultiByte(CP_UTF8, 0, (wchar_t *)pvValue, -1, c, ulLength * 2, NULL, NULL);
				result = str.assign(c);
				delete[] c;
			}
			break;
			case DBTYPE_IUNKNOWN:
				// We've bound this as an ISequentialStream object,
				// therefore the data in our buffer is a pointer
				// to the object's ISequentialStream interface

				pISeqStream = *(ISequentialStream**)pvValue;

				// We call ISequentialStream::Read to read bytes from
				// the stream blindly into our buffer, simply as a
				// demonstration of ISequentialStream. To display the
				// data properly, the native provider type of this
				// column should be accounted for; it could be
				// DBTYPE_WSTR, in which case this works, or it could
				// be DBTYPE_STR or DBTYPE_BYTES, in which case this
				// won't display the data correctly

				// Isto aqui est errado pois nao est convertendo pra UTF8. 
				longValueBuf = NULL;
				do {
					hr = pISeqStream->Read(pbBuff, cbNeeded, &cbRead);
					longValueBuf = (char *)CoTaskMemRealloc(longValueBuf, cbReadTotal + cbRead + 2);
					memcpy((void*)&longValueBuf[cbReadTotal], pbBuff, cbRead);
					cbReadTotal += cbRead;
				}
				while (SUCCEEDED(hr) && hr != S_FALSE && cbRead == cbNeeded);

				if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

				// Since streams don't provide NULL-termination,
				// we'll NULL-terminate the resulting string ourselves
				longValueBuf[cbReadTotal] = '\0'; longValueBuf[cbReadTotal+1] = '\0';

				c = new char[cbReadTotal * 2];
				WideCharToMultiByte(CP_UTF8, 0, (wchar_t *)longValueBuf, -1, c, cbReadTotal * 2, NULL, NULL);
				result = str.assign(c);
				delete[] c;
				CoTaskMemFree(longValueBuf);

				// Release the stream object, now that we're done
				pISeqStream->Release();
				pISeqStream = NULL;
			break;

			// This is an error status, so don't try to display the data
		default:
			result = nullptr;
			break;
		}
	}
CLEANUP:
	if (pISeqStream) {
		pISeqStream->Release();
	}
	return result;
}

json rowToJsonArray(DBORDINAL cBindings, DBBINDING *rgBindings, DBCOLUMNINFO *rgColumnInfo, void *pData, int* errorCount, std::string* errorMessages) {
	ULONG				iCol;

	json jsonResult;

	// For each column that we have bound, display the data
	for (iCol = 0; iCol < cBindings; iCol++) {
		jsonResult[iCol] = dbColumnValueToUF8(rgBindings, rgColumnInfo, pData, iCol, errorCount, errorMessages);
	}
	return jsonResult;
}

json sqlExecute(IDBInitialize* pInitialize, const LPCOLESTR sqlCommand, int* errorCount, std::string* errorMessages) {
	IUnknown*			pSession = NULL; 
	IDBCreateSession*	pIDBCreateSession = NULL;
	IDBCreateCommand*	pIDBCreateCommand = NULL;
	ICommandProperties* pICommandProperties = NULL;
	ICommandText*		pICommandText = NULL;
	IMultipleResults*	pIMultipleResults = NULL;
	IRowset*			pIRowset = NULL;
	HRESULT				hr = S_OK;
	DBORDINAL			cBindings = 0;
	DBBINDING*			rgBindings = NULL;
	DBCOLUMNINFO*       rgColumnInfo = NULL;
	LPWSTR				pStringBuffer = NULL;
	void*				pData = NULL;
	void*				pCurData = NULL;
	DBCOUNTITEM			iRow = 0;
	DBROWCOUNT			cRowsAffected;
	//DBPROPSET			rgCmdPropSet[1];
	//DBPROP			rgCmdProperties[1];
	json				jsonColumns = json::array();
	json				jsonRows = json::array();
	json				jsonResult = json::object();

	hr = pInitialize->QueryInterface(IID_IDBCreateSession, (void**)&pIDBCreateSession);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	hr = pIDBCreateSession->CreateSession(NULL, IID_IDBCreateCommand, &pSession);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	hr = pSession->QueryInterface(IID_IDBCreateCommand, (void**)&pIDBCreateCommand);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	hr = pIDBCreateCommand->CreateCommand(NULL,	IID_ICommandText, (IUnknown**)&pICommandText);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	hr = pICommandText->SetCommandText(DBGUID_DBSQL, sqlCommand);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	/*Cdigo necessrio para "long Values", p. ex: varchar(max). Acabou se mostrando desnecessrio nos testes
	hr = pICommandText->QueryInterface(IID_ICommandProperties, (void**)&pICommandProperties);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	rgCmdProperties[0].dwPropertyID = DBPROP_ACCESSORDER;
	rgCmdProperties[0].vValue.vt = VT_I4;
	rgCmdProperties[0].vValue.lVal = DBPROPVAL_AO_SEQUENTIAL;
	rgCmdProperties[0].dwOptions = DBPROPOPTIONS_REQUIRED;
	rgCmdProperties[0].colid = DB_NULLID;

	rgCmdPropSet[0].guidPropertySet = DBPROPSET_ROWSET;
	rgCmdPropSet[0].cProperties = 1;
	rgCmdPropSet[0].rgProperties = rgCmdProperties;

	hr = pICommandProperties->SetProperties(1, rgCmdPropSet);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;*/

	hr = pICommandText->Execute(NULL, IID_IRowset, NULL, &cRowsAffected, (IUnknown**)&pIRowset);
	if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

	jsonResult["rowsAffected"] = cRowsAffected;
	
	if (pIRowset != NULL) {
		HACCESSOR	hAccessor = DB_NULL_HACCESSOR;
		DBORDINAL	cbRowSize;
		DBCOUNTITEM cRowsReturned, totalRows = 0;
		HROW*		pRow = NULL;

		hr = createAccessor(pIRowset, &hAccessor, &cBindings, &rgBindings, &rgColumnInfo, &pStringBuffer, &cbRowSize, errorCount, errorMessages);
		if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

		for (DBORDINAL iCol = 0; iCol < cBindings; iCol++) {
			jsonColumns[iCol] = { { "name", WideCharToUTF8(rgColumnInfo[iCol].pwszName) }, {"type", rgColumnInfo[iCol].wType} };
		}

		// Allocate enough memory to hold MAX_ROWS rows of data; this is
		// where the actual row data from the provider will be placed
		pData = CoTaskMemAlloc(cbRowSize * MAX_ROWS);
		if (memoryCheck(hr, pData, errorCount, errorMessages) < 0) goto CLEANUP;

		while (hr == S_OK) {
			// Attempt to get MAX_ROWS row handles from the provider
			hr = pIRowset->GetNextRows(DB_NULL_HCHAPTER, 0, MAX_ROWS, &cRowsReturned, &pRow);
			if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;

			// get data into pData;
			for (iRow = 0; iRow < cRowsReturned; iRow++) {
				pCurData = (BYTE*)pData + (cbRowSize * iRow);
				hr = pIRowset->GetData(pRow[iRow], hAccessor, pCurData);
				//if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;
			}

			// moving to dartRows
			for (iRow = 0; iRow < cRowsReturned; iRow++) {
				pCurData = (BYTE*)pData + (cbRowSize * iRow);
				jsonRows[totalRows + iRow] = rowToJsonArray(cBindings, rgBindings, rgColumnInfo, pCurData, errorCount, errorMessages);
			}

			// release row handles and alloc dartRows array
			if (cRowsReturned)
			{
				hr = pIRowset->ReleaseRows(cRowsReturned, pRow, NULL, NULL, NULL);
				if (oleCheck(hr, errorCount, errorMessages) < 0) goto CLEANUP;
				CoTaskMemFree(pRow);
				pRow = NULL;
			}

			// updating totalRows with all rows returned until now
			totalRows += cRowsReturned;
		}
	}

CLEANUP:
	if (pICommandProperties) {
		pICommandProperties->Release();
		pICommandProperties = NULL;
	}
	if (pIRowset != NULL) {
		pIRowset->Release();
	}
	if (pICommandText != NULL) {
		pICommandText->Release();
	}
	if (pIDBCreateCommand != NULL) {
		pIDBCreateCommand->Release();
	}
	if (pSession != NULL) {
		pSession->Release();
	}
	if (pIDBCreateSession != NULL) {
		pIDBCreateSession->Release();
	}

	CoTaskMemFree(pStringBuffer);
	CoTaskMemFree(rgColumnInfo);
	freeBindings(cBindings, rgBindings);
	CoTaskMemFree(pData);
	jsonResult["columns"] = jsonColumns;
	jsonResult["rows"] = jsonRows;
	return jsonResult;
};

