import 'dart:async';
import 'dart:convert';
import 'dart:isolate';
import 'dart-ext:dart_mssql';

import 'sql_result.dart';

class MssqlConnection {
  String _username;
  String _password;
  String _host;
  String _databaseName;

  static SendPort __executePort;

  MssqlConnection(
      this._host, this._databaseName, this._username, this._password);

  Future _executeCommand(String sqlCommand) {
    final completer = Completer();
    final replyPort = RawReceivePort();
    final args = List(7);
    args[0] = replyPort.sendPort;
    args[1] = _host;
    args[2] = _databaseName;
    args[3] = _username;
    args[4] = _password;
    args[5] = 0; // authType por enquanto é só 0 (user/password)
    args[6] = sqlCommand;

    _executePort.send(args);
    replyPort.handler = (result) {
      replyPort.close();
      if (result == null) {
        throw "Unknown error";
      } else {
        completer.complete(json.decode(result));
      }
    };
    return completer.future;
  }

  SendPort get _executePort {
    return __executePort ?? _newExecuteServicePort();
  }

  SendPort _newExecuteServicePort() native "execute_ServicePort";

  Future<SqlResult> execute(String sqlCommand) async {
    Map result = await _executeCommand(sqlCommand);
    if (result.containsKey("error")) {
      throw result["error"];
    }
    return SqlResult(
        columns: result["columns"],
        rows: result["rows"],
        rowsAffected: result["rowsAffected"],
        lastIdentity: result["lastIdentity"]);
  }
}
