import 'package:quiver/strings.dart';

String strSlice(String s, [int num = 1]) {
  return isEmpty(s) ? "" : s.substring(0, s.length - 1);
}

String sqlDate(DateTime d) {
  String r = d.toString();
  r = r.substring(0, r.length-4);
  if (r.endsWith(".")) {
    r += "0";
  }
  return r;
}

class SqlResult {
  List columns;
  List rows;
  int rowsAffected = -1;
  int rowIndex = 0;
  int lastIdentity = -1;
  String error;
  Map<String, dynamic> fieldIndexes = Map();

  void updateFieldIndexes() {
    int i = 0;
    columns?.forEach((c) {
      fieldIndexes[c["name"]] = i;
      i++;
    });
  }

  SqlResult({this.columns, this.rows, this.rowsAffected, this.lastIdentity}) {
    rowIndex = 0;
    updateFieldIndexes();
  }

  SqlResult.fromMap(Map map) {
    this.rows = map["rows"];
    this.columns = map["columns"];
    this.rowsAffected = map["rowsAffected"];
    this.lastIdentity = map["lastIdentity"];
    this.fieldIndexes = map["fieldIndexes"];
    this.error = map["error"];
  }

  int index(String fieldName) {
    return fieldIndexes[fieldName];
  }

  String values(String fieldName) {
    int fieldIndex = fieldIndexes[fieldName];
    String ids = "-1,";
    rows.forEach((e) {
      ids += e[fieldIndex].toString() + ",";
    });
    ids = strSlice(ids);
    return ids;
  }

  dynamic value(String fieldName) {
    if (rowIndex < 0 || rowIndex >= rows.length) {
      return null;
    } else {
      return rows[rowIndex][fieldIndexes[fieldName]];
    }
  }

  Map<String, dynamic> map() {
    Map<String, dynamic> result = Map();
    if (rowIndex < 0 || rowIndex >= rows.length) {
      columns.forEach((c) {
        result[c["name"]] = null;
      });
    } else {
      columns.forEach((c) {
        result[c["name"]] = rows[rowIndex][fieldIndexes[c["name"]]];
      });
    }
    return result;
  }

  bool next() {
    return (rows == null || rowIndex == rows.length)
        ? false
        : () {
            rowIndex++;
            return true;
          };
  }

  bool prior() {
    return (rows == null || rowIndex == 0)
        ? false
        : () {
            rowIndex--;
            return true;
          };
  }

  Map asMap() {
    return {
      "columns": columns,
      "rows": rows,
      "rowsAffected": rowsAffected,
      "lastIdentity": lastIdentity,
      "fieldIndexes": fieldIndexes,
      "error": error,
    };
  }

  String mapToInsert(String tableName, Map<String, dynamic> map, [String fieldList]) {
    String fieldNames = "";
    String fieldValues = "";
    if (fieldList != null) {
      fieldList = ",$fieldList,";
    }
    map.forEach((n, v) {
      if (fieldList == null || fieldList.contains(",$n,")) {
        fieldNames += n + ',';
        if (v is String) {
          if (n.indexOf("dat_") == 0) {
            fieldValues +=
                "${v == null ? null : "'" + sqlDate(DateTime.tryParse(v).toUtc()) + "'"},";
          } else {
            fieldValues +=
                "${v == null ? null : "'" + v.replaceAll("'", "''") + "'"},";
          }
        } else {
          fieldValues += "$v,";
        }
      }
    });
    fieldNames = strSlice(fieldNames);
    fieldValues = strSlice(fieldValues);
    return "insert into $tableName($fieldNames) values ($fieldValues)";
  }

  String mapToUpdate(
      String tableName, Map<String, dynamic> map, String whereClause) {
    String fieldValues = "";
    map.forEach((n, v) {
      if (v is String) {
          if (n.indexOf("dat_") == 0) {
            fieldValues +=
              "$n = ${v == null ? null : "'" + sqlDate(DateTime.tryParse(v).toUtc()) + "'"},";
          } else {
            fieldValues +=
              "$n = ${v == null ? null : "'" + v.replaceAll("'", "''") + "'"},";
          }
      } else {
        fieldValues += "$n = $v,";
      }
    });
    fieldValues = strSlice(fieldValues);
    return "update $tableName set $fieldValues where $whereClause";
  }
}
